import pandas as pd

def describe_outliers(df, threshold=1.5, numeric=True):
    """
    Extends the functionality of pandas.Dataframe.describe() for numeric data by 
    additionally providing a count of lower-tail and upper-tail outliers. 

    If the data contains numerical values the printed description includes the following
    information:
    
    - dtype: the datatype of the column. 
    - Non-null count: The number of non-null observations.
    - mean: The mean value.
    - standard deviation: The standard deviation.
    - min: the minimum value.
    - 25% percentile: The 25% percentile (Q1).
    - 50% (median): The 50% percentile (Q2).
    - 75% percentile: The 75% percentile (Q3).
    - max: the maximum value.
    - lower-tail outliers: count of values significantly smaller than the majority of data.
    - upper-tail outliers: count of values significantly larger than the majority of data.

    The number of outliers is calculated using the Interquartile Range (IQR) 
    method for outlier detection. Lower-tail outliers are observations less than:
    Q1 - threshold*IQR. Upper-tail outliers are observations greater than: Q3 + threshold*IQR. 

    Parameters
    ----------
    df : pandas.DataFrame
        A tidy dataframe containing at least 1 numeric column. 
    
    threshold : float, optional
        The scalar used in outlier detection. It must be a non-negative numeric value. A higher 
        value reduces the sensitivity of outlier detection. Default is 1.5 
    
    numeric : boolean, optional
        If set to True, the returned dataframe will only include description on numeric columns. 
        If False, the output will include the dtype and count for non-numeric columns. Default is True.  

    Returns
    -------
    summary_df : pandas.DataFrame
        A dataframe of the descriptive summary statistics. 
    
    Example
    --------
    >>>  import pandas as pd
    >>>  data = {'numeric': [1, 2, 3, 4, 5, 100], 
    >>>         'categorical': ['a', 'b', 'c', 'd', 'e', 'f']}
    >>>  df = pd.DataFrame(data)
    >>>  describe_outliers(df, threshold = 2, numeric=False)
    # Output 
                        categorical	    numeric
        dtype	            object	    int64
        Non-null count	    6	        6
        mean	            NaN	        19.166667
        standard deviation	NaN	        39.625329
        min value	        NaN	        1.0
        25% percentile	    NaN	        2.25
        50% (median)	    NaN	        3.5
        75% percentile	    NaN	        4.75
        max value	        NaN	        100.0
        lower-tail outliers	NaN	        0.0
        upper-tail outliers	NaN	        1.0
    """
    
    if not isinstance(df, pd.DataFrame):
        raise TypeError("Input df must be a DataFrame.")

    if not threshold >= 0:
        raise ValueError("Invalid value for threshold. Threshold must be a non-negative number.")
    
    column_names = df.columns
    numeric_columns = df.select_dtypes(include='number').columns.tolist()

    # consider only numeric columns (unless specified)
    if numeric == True:
        column_names = numeric_columns

    if len(numeric_columns) == 0:
        raise ValueError("Your dataframe contains no numeric columns. It should include at least 1.")
    
    # calculate summary statistics
    counts = df[column_names].count().astype(int)
    mean = df[numeric_columns].mean()
    sd = df[numeric_columns].std()
    min = pd.Series(df[numeric_columns].min(), index=numeric_columns)
    q1 = df[numeric_columns].quantile(0.25)
    q2 = df[numeric_columns].quantile(0.50)
    q3 = df[numeric_columns].quantile(0.75)
    max = pd.Series(df[numeric_columns].max(), index=numeric_columns)

    # outlier detection
    iqr = q3 - q1
    lower_fences = q1 - threshold*iqr
    upper_fences = q3 + threshold*iqr
    lower_outliers_count = (df[numeric_columns] < lower_fences).sum()
    upper_outliers_count = (df[numeric_columns] > upper_fences).sum()

    # display the description
    summary_df = pd.DataFrame({
        'dtype': df.dtypes[column_names],
        'Non-null count': counts,
        'mean': mean,
        'standard deviation': sd,
        'min value': min,
        '25% percentile': q1,
        '50% (median)': q2,
        '75% percentile': q3,
        'max value': max,
        'lower-tail outliers': lower_outliers_count,
        'upper-tail outliers': upper_outliers_count
    }).T
    
    return summary_df


    


import pandas as pd
import numpy as np


def cor_eda(dataset, na_handling="drop"):
    """
    Perform exploratory data analysis (EDA) by calculating the correlation between numerical variables.
    This function isolates the numerically variables from the given dataset and handles NA values,
    calculates the correlation between each pair of variables, and displays the results in a table format.

    Parameters:
    dataset (DataFrame): The DataFrame should include various types of variables
    na_handling (str, optional): Specifies the method to handle missing values (NAs). Options are 'drop', 'mean', 'median', and 'value'. Default is 'drop'.
        - 'drop': Drops rows with any NAs.
        - 'mean': Replaces NAs with the mean value of the column.
        - 'median': Replaces NAs with the median value of the column.
        - 'value': Replaces NAs with a selected value.

    Returns:
    DataFrame: A DataFrame containing the correlation values between each pair of numerical variables.

    Example:
    cor_eda(data, na_handling='mean')
            age    salary
    age     1.0000   0.9769
    salary  0.9769   1.0000

    """
    # Isolate the numerical variables
    numerical_data = dataset.select_dtypes(include=["number"])

    if numerical_data.empty or numerical_data.shape[1] == 0:
        return "There are no numerical columns"

    # Handle missing values according to the specified method
    if na_handling == "drop":
        numerical_data = numerical_data.dropna()
    elif na_handling == "mean":
        numerical_data = numerical_data.fillna(numerical_data.mean())
    elif na_handling == "median":
        numerical_data = numerical_data.fillna(numerical_data.median())
    elif na_handling == "value":
        # Replace with a chosen value, for demonstration we use 0
        numerical_data = numerical_data.fillna(0)
    else:
        raise ValueError("na_handling must be 'drop', 'mean', 'median', or 'value'")

    # Use pandas built-in corr() method to get the correlation matrix
    correlation_matrix = numerical_data.corr()

    return correlation_matrix.astype(float)



import numpy as np
import pandas as pd
import warnings


def info_na(df):
    """
    This function replicates and extends behaviour of pandas.DataFrame.info(), 
    which provides type, shape, memory useage, and column information.
    New information will consist of row-level summary statistics for null values to characterize dataframe structure.

    This function prints the following information about a DataFrame:
    (starred information exists in pandas.DataFrame.info())
    - DataFrame Type *
    - Shape *
    - Memory Usage *
    - Columns:
        - Index *
        - Name *
        - Null Count
        - Null Percentage
        - Dtype *
    - Rows:
        - Total Rows:
        - Any Null Count: Count of rows with any Null Values.
        - Any Null Percent: Percentage of rows with any Null Values.
        - All Null Count: Count of rows with all Null Values.
        - All Null Percent: Percentage of rows with all Null Values.
        - Mean Null Count: Average number of Null Values per row.
        - Std.Dev Null Count: Standard devation of the number of Null Values per row.
        - Max Null Count: The maximum number of Null Values found in a row.
        - Min Null Count: The minimum number of Null Values found in a row.

    Parameters
    ----------
    df : pandas.DataFrame
        A tidy dataframe.

    Returns
    -------
    None
        Descriptive information about the dataframe will be printed to the console. 

    Examples
    --------
    >>> info_na(pd.DataFrame(
            [
                [np.nan, 13, "hello"],
                [np.nan, np.nan, "this"],
                [37, 45, "is"],
                [256, 31, ""],
                [1, np.nan, "test"],
            ],
            index=["First", "Second", "Third", "Fourth", "Fifth"],
            columns=["Column1", "ColumnNumber2", "Column3"],
        ))
    type: <class 'pandas.core.frame.DataFrame'>
    shape: (5, 3)
    memory usage: 692 B
    --------
    columns:
    #        column  null count  null %   dtype
    0       Column1           2    40.0 float64
    1 ColumnNumber2           2    40.0 float64
    2       Column3           0     0.0  object
    -----
    rows:
    total rows             5.00
    any null count         3.00
    any null %            60.00
    all null count         0.00
    all null %             0.00
    mean null count        0.80
    std.dev null count     0.84
    max null count         2.00
    min null count         0.00
    """

    # Input checks and warnings
    if not isinstance(df, pd.DataFrame):
        raise TypeError("Input `df` must be a Pandas DataFrame")

    if pd.isna(df).all(axis=None):
        warnings.warn("Input `df` contains all NA values")

    # Collect information about df
    type_info = type(df)

    shape_info = df.shape

    column_info = pd.DataFrame(
        {
            "#": np.arange(df.shape[1]),
            "column": df.columns.values,
            "null count": df.isna().sum(axis=0),
            "null %": (df.isna().sum(axis=0) / df.shape[0] * 100).round(2),
            "dtype": df.dtypes,
        }
    )

    row_info = pd.Series(
        {
            "total rows": df.shape[0],
            "any null count": df.isna().any(axis=1).sum(),
            "any null %": (df.isna().any(axis=1).sum() / df.shape[0] * 100).round(2),
            "all null count": df.isna().all(axis=1).sum(),
            "all null %": (df.isna().all(axis=1).sum() / df.shape[0] * 100).round(2),
            "mean null count": df.isna().sum(axis=1).mean().round(2),
            "std.dev null count": df.isna().sum(axis=1).std().round(2),
            "max null count": df.isna().sum(axis=1).max(),
            "min null count": df.isna().sum(axis=1).min(),
        }
    )

    # Human-readable memory usage formatting
    suffix = ["B", "KB", "MB", "GB", "TB"]
    memory_bytes = df.memory_usage(deep=True).sum()
    n = 0
    while memory_bytes > 2**10:
        memory_bytes = memory_bytes / 2**10
        n += 1
    memory_info = f"{np.round(memory_bytes, 1)} {suffix[n]}"

    # Format to output string
    output = f"""
type: {type_info}
shape: {shape_info}
memory usage: {memory_info}
--------
columns:
{column_info.to_string(index=False)}
-----
rows:
{row_info.to_string()}
"""

    print(output)
